@@ -1,5 +1,5 @@
 /* cp.c  -- file copying (main routines)
-   Copyright (C) 1989-1991, 1995-2011 Free Software Foundation, Inc.
+   Copyright (C) 1989-2023 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
    Written by Torbjorn Granlund, David MacKenzie, and Jim Meyering. */
 
@@ -20,40 +20,28 @@
 #include <stdio.h>
 #include <sys/types.h>
 #include <getopt.h>
-#include <selinux/selinux.h>
+#include <selinux/label.h>
 
 #include "system.h"
 #include "argmatch.h"
 #include "backupfile.h"
 #include "copy.h"
 #include "cp-hash.h"
+#include "die.h"
 #include "error.h"
 #include "filenamecat.h"
 #include "ignore-value.h"
 #include "quote.h"
 #include "stat-time.h"
+#include "targetdir.h"
 #include "utimens.h"
 #include "acl.h"
 
-#if ! HAVE_LCHOWN
-# define lchown(name, uid, gid) chown (name, uid, gid)
-#endif
-
-#define ASSIGN_BASENAME_STRDUPA(Dest, File_name)	\
-  do							\
-    {							\
-      char *tmp_abns_;					\
-      ASSIGN_STRDUPA (tmp_abns_, (File_name));		\
-      Dest = last_component (tmp_abns_);		\
-      strip_trailing_slashes (Dest);			\
-    }							\
-  while (0)
-
-/* The official name of this program (e.g., no `g' prefix).  */
+/* The official name of this program (e.g., no 'g' prefix).  */
 #define PROGRAM_NAME "cp"
 
 #define AUTHORS \
-  proper_name_utf8 ("Torbjorn Granlund", "Torbj\303\266rn Granlund"), \
+  proper_name ("Torbjorn Granlund"), \
   proper_name ("David MacKenzie"), \
   proper_name ("Jim Meyering")
 
@@ -74,6 +62,7 @@ enum
 {
   ATTRIBUTES_ONLY_OPTION = CHAR_MAX + 1,
   COPY_CONTENTS_OPTION,
+  DEBUG_OPTION,
   NO_PRESERVE_ATTRIBUTES_OPTION,
   PARENTS_OPTION,
   PRESERVE_ATTRIBUTES_OPTION,
@@ -105,11 +94,11 @@ ARGMATCH_VERIFY (sparse_type_string, sparse_type);
 
 static char const *const reflink_type_string[] =
 {
-  "auto", "always", NULL
+  "auto", "always", "never", NULL
 };
 static enum Reflink_type const reflink_type[] =
 {
-  REFLINK_AUTO, REFLINK_ALWAYS
+  REFLINK_AUTO, REFLINK_ALWAYS, REFLINK_NEVER
 };
 ARGMATCH_VERIFY (reflink_type_string, reflink_type);
 
@@ -119,6 +108,7 @@ static struct option const long_opts[] =
   {"attributes-only", no_argument, NULL, ATTRIBUTES_ONLY_OPTION},
   {"backup", optional_argument, NULL, 'b'},
   {"copy-contents", no_argument, NULL, COPY_CONTENTS_OPTION},
+  {"debug", no_argument, NULL, DEBUG_OPTION},
   {"dereference", no_argument, NULL, 'L'},
   {"force", no_argument, NULL, 'f'},
   {"interactive", no_argument, NULL, 'i'},
@@ -141,6 +131,7 @@ static struct option const long_opts[] =
   {"target-directory", required_argument, NULL, 't'},
   {"update", no_argument, NULL, 'u'},
   {"verbose", no_argument, NULL, 'v'},
+  {GETOPT_SELINUX_CONTEXT_OPTION_DECL},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
@@ -150,8 +141,7 @@ void
 usage (int status)
 {
   if (status != EXIT_SUCCESS)
-    fprintf (stderr, _("Try `%s --help' for more information.\n"),
-             program_name);
+    emit_try_help ();
   else
     {
       printf (_("\
@@ -162,11 +152,10 @@ Usage: %s [OPTION]... [-T] SOURCE DEST\n\
               program_name, program_name, program_name);
       fputs (_("\
 Copy SOURCE to DEST, or multiple SOURCE(s) to DIRECTORY.\n\
-\n\
-"), stdout);
-      fputs (_("\
-Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
+
+      emit_mandatory_arg_note ();
+
       fputs (_("\
   -a, --archive                same as -dR --preserve=all\n\
       --attributes-only        don't copy the file data, just the attributes\n\
@@ -175,19 +164,21 @@ Mandatory arguments to long options are mandatory for short options too.\n\
   -b                           like --backup but does not accept an argument\n\
       --copy-contents          copy contents of special files when recursive\n\
   -d                           same as --no-dereference --preserve=links\n\
+"), stdout);
+      fputs (_("\
+      --debug                  explain how a file is copied.  Implies -v\n\
 "), stdout);
       fputs (_("\
   -f, --force                  if an existing destination file cannot be\n\
-                                 opened, remove it and try again (redundant if\
-\n\
-                                 the -n option is used)\n\
+                                 opened, remove it and try again (this option\n\
+                                 is ignored when the -n option is also used)\n\
   -i, --interactive            prompt before overwrite (overrides a previous -n\
 \n\
                                   option)\n\
   -H                           follow command-line symbolic links in SOURCE\n\
 "), stdout);
       fputs (_("\
-  -l, --link                   link files instead of copying\n\
+  -l, --link                   hard link files instead of copying\n\
   -L, --dereference            always follow symbolic links in SOURCE\n\
 "), stdout);
       fputs (_("\
@@ -197,11 +188,7 @@ Mandatory arguments to long options are mandatory for short options too.\n\
 "), stdout);
       fputs (_("\
   -p                           same as --preserve=mode,ownership,timestamps\n\
-      --preserve[=ATTR_LIST]   preserve the specified attributes (default:\n\
-                                 mode,ownership,timestamps), if possible\n\
-                                 additional attributes: context, links, xattr,\
-\n\
-                                 all\n\
+      --preserve[=ATTR_LIST]   preserve the specified attributes\n\
 "), stdout);
       fputs (_("\
       --no-preserve=ATTR_LIST  don't preserve the specified attributes\n\
@@ -230,75 +217,77 @@ Mandatory arguments to long options are mandatory for short options too.\n\
                                  destination file is missing\n\
   -v, --verbose                explain what is being done\n\
   -x, --one-file-system        stay on this file system\n\
+"), stdout);
+      fputs (_("\
+  -Z                           set SELinux security context of destination\n\
+                                 file to default type\n\
+      --context[=CTX]          like -Z, or if CTX is specified then set the\n\
+                                 SELinux or SMACK security context to CTX\n\
 "), stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
       fputs (_("\
 \n\
+ATTR_LIST is a comma-separated list of attributes. Attributes are 'mode' for\n\
+permissions (including any ACL and xattr permissions), 'ownership' for user\n\
+and group, 'timestamps' for file timestamps, 'links' for hard links, 'context'\
+\nfor security context, 'xattr' for extended attributes, and 'all' for all\n\
+attributes.\n\
+"), stdout);
+      fputs (_("\
+\n\
 By default, sparse SOURCE files are detected by a crude heuristic and the\n\
 corresponding DEST file is made sparse as well.  That is the behavior\n\
 selected by --sparse=auto.  Specify --sparse=always to create a sparse DEST\n\
 file whenever the SOURCE file contains a long enough sequence of zero bytes.\n\
 Use --sparse=never to inhibit creation of sparse files.\n\
+"), stdout);
+      fputs (_("\
 \n\
 When --reflink[=always] is specified, perform a lightweight copy, where the\n\
 data blocks are copied only when modified.  If this is not possible the copy\n\
 fails, or if --reflink=auto is specified, fall back to a standard copy.\n\
+Use --reflink=never to ensure a standard copy is performed.\n\
 "), stdout);
-      fputs (_("\
-\n\
-The backup suffix is `~', unless set with --suffix or SIMPLE_BACKUP_SUFFIX.\n\
-The version control method may be selected via the --backup option or through\n\
-the VERSION_CONTROL environment variable.  Here are the values:\n\
-\n\
-"), stdout);
-      fputs (_("\
-  none, off       never make backups (even if --backup is given)\n\
-  numbered, t     make numbered backups\n\
-  existing, nil   numbered if numbered backups exist, simple otherwise\n\
-  simple, never   always make simple backups\n\
-"), stdout);
+      emit_backup_suffix_note ();
       fputs (_("\
 \n\
 As a special case, cp makes a backup of SOURCE when the force and backup\n\
 options are given and SOURCE and DEST are the same name for an existing,\n\
 regular file.\n\
 "), stdout);
-      emit_ancillary_info ();
+      emit_ancillary_info (PROGRAM_NAME);
     }
   exit (status);
 }
 
-/* Ensure that the parent directories of CONST_DST_NAME have the
+/* Ensure that parents of CONST_DST_NAME aka DST_DIRFD+DST_RELNAME have the
    correct protections, for the --parents option.  This is done
    after all copying has been completed, to allow permissions
    that don't include user write/execute.
 
-   SRC_OFFSET is the index in CONST_DST_NAME of the beginning of the
-   source directory name.
-
    ATTR_LIST is a null-terminated linked list of structures that
    indicates the end of the filename of each intermediate directory
    in CONST_DST_NAME that may need to have its attributes changed.
-   The command `cp --parents --preserve a/b/c d/e_dir' changes the
+   The command 'cp --parents --preserve a/b/c d/e_dir' changes the
    attributes of the directories d/e_dir/a and d/e_dir/a/b to match
    the corresponding source directories regardless of whether they
-   existed before the `cp' command was given.
+   existed before the 'cp' command was given.
 
    Return true if the parent of CONST_DST_NAME and any intermediate
    directories specified by ATTR_LIST have the proper permissions
    when done.  */
 
 static bool
-re_protect (char const *const_dst_name, size_t src_offset,
+re_protect (char const *const_dst_name, int dst_dirfd, char const *dst_relname,
             struct dir_attr *attr_list, const struct cp_options *x)
 {
   struct dir_attr *p;
   char *dst_name;		/* A copy of CONST_DST_NAME we can change. */
-  char *src_name;		/* The source name in `dst_name'. */
+  char *src_name;		/* The source name in 'dst_name'. */
 
   ASSIGN_STRDUPA (dst_name, const_dst_name);
-  src_name = dst_name + src_offset;
+  src_name = dst_name + (dst_relname - const_dst_name);
 
   for (p = attr_list; p; p = p->next)
     {
@@ -315,27 +304,27 @@ re_protect (char const *const_dst_name, size_t src_offset,
           timespec[0] = get_stat_atime (&p->st);
           timespec[1] = get_stat_mtime (&p->st);
 
-          if (utimens (dst_name, timespec))
+          if (utimensat (dst_dirfd, src_name, timespec, 0))
             {
               error (0, errno, _("failed to preserve times for %s"),
-                     quote (dst_name));
+                     quoteaf (dst_name));
               return false;
             }
         }
 
       if (x->preserve_ownership)
         {
-          if (lchown (dst_name, p->st.st_uid, p->st.st_gid) != 0)
+          if (lchownat (dst_dirfd, src_name, p->st.st_uid, p->st.st_gid) != 0)
             {
               if (! chown_failure_ok (x))
                 {
                   error (0, errno, _("failed to preserve ownership for %s"),
-                         quote (dst_name));
+                         quoteaf (dst_name));
                   return false;
                 }
               /* Failing to preserve ownership is OK. Still, try to preserve
                  the group, but ignore the possible error. */
-              ignore_value (lchown (dst_name, -1, p->st.st_gid));
+              ignore_value (lchownat (dst_dirfd, src_name, -1, p->st.st_gid));
             }
         }
 
@@ -346,10 +335,10 @@ re_protect (char const *const_dst_name, size_t src_offset,
         }
       else if (p->restore_mode)
         {
-          if (lchmod (dst_name, p->st.st_mode) != 0)
+          if (lchmodat (dst_dirfd, src_name, p->st.st_mode) != 0)
             {
               error (0, errno, _("failed to preserve permissions for %s"),
-                     quote (dst_name));
+                     quoteaf (dst_name));
               return false;
             }
         }
@@ -365,6 +354,7 @@ re_protect (char const *const_dst_name, size_t src_offset,
    SRC_OFFSET is the index in CONST_DIR (which is a destination
    directory) of the beginning of the source directory name.
    Create any leading directories that don't already exist.
+   DST_DIRFD is a file descriptor for the target directory.
    If VERBOSE_FMT_STRING is nonzero, use it as a printf format
    string for printing a message after successfully making a directory.
    The format should take two string arguments: the names of the
@@ -380,6 +370,7 @@ re_protect (char const *const_dst_name, size_t src_offset,
 
 static bool
 make_dir_parents_private (char const *const_dir, size_t src_offset,
+                          int dst_dirfd,
                           char const *verbose_fmt_string,
                           struct dir_attr **attr_list, bool *new_dst,
                           const struct cp_options *x)
@@ -388,20 +379,29 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
   char *dir;		/* A copy of CONST_DIR we can change.  */
   char *src;		/* Source name in DIR.  */
   char *dst_dir;	/* Leading directory of DIR.  */
-  size_t dirlen;	/* Length of DIR.  */
+  idx_t dirlen = dir_len (const_dir);
+
+  *attr_list = NULL;
+
+  /* Succeed immediately if the parent of CONST_DIR must already exist,
+     as the target directory has already been checked.  */
+  if (dirlen <= src_offset)
+    return true;
 
   ASSIGN_STRDUPA (dir, const_dir);
 
   src = dir + src_offset;
 
-  dirlen = dir_len (dir);
   dst_dir = alloca (dirlen + 1);
   memcpy (dst_dir, dir, dirlen);
   dst_dir[dirlen] = '\0';
+  char const *dst_reldir = dst_dir + src_offset;
+  while (*dst_reldir == '/')
+    dst_reldir++;
 
-  *attr_list = NULL;
-
-  if (stat (dst_dir, &stats) != 0)
+  /* XXX: If all dirs are present at the destination,
+     no permissions or security contexts will be updated.  */
+  if (fstatat (dst_dirfd, dst_reldir, &stats, 0) != 0)
     {
       /* A parent of CONST_DIR does not exist.
          Make all missing intermediate directories. */
@@ -410,13 +410,15 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
       slash = src;
       while (*slash == '/')
         slash++;
+      dst_reldir = slash;
+
       while ((slash = strchr (slash, '/')))
         {
-          struct dir_attr *new IF_LINT ( = NULL);
+          struct dir_attr *new;
           bool missing_dir;
 
           *slash = '\0';
-          missing_dir = (stat (dir, &stats) != 0);
+          missing_dir = fstatat (dst_dirfd, dst_reldir, &stats, 0) != 0;
 
           if (missing_dir || x->preserve_ownership || x->preserve_mode
               || x->preserve_timestamps)
@@ -432,7 +434,7 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
               if (src_errno)
                 {
                   error (0, src_errno, _("failed to get attributes of %s"),
-                         quote (src));
+                         quoteaf (src));
                   return false;
                 }
 
@@ -444,6 +446,12 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
               *attr_list = new;
             }
 
+          /* If required set the default context for created dirs.  */
+          if (! set_process_security_ctx (src, dir,
+                                          missing_dir ? new->st.st_mode : 0,
+                                          missing_dir, x))
+            return false;
+
           if (missing_dir)
             {
               mode_t src_mode;
@@ -452,7 +460,7 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
 
               /* This component does not exist.  We must set
                  *new_dst and new->st.st_mode inside this loop because,
-                 for example, in the command `cp --parents ../a/../b/c e_dir',
+                 for example, in the command 'cp --parents ../a/../b/c e_dir',
                  make_dir_parents_private creates only e_dir/../a if
                  ./b already exists. */
               *new_dst = true;
@@ -472,11 +480,12 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
                  (src_mode & ~S_IRWXUGO) != 0.  However, common practice is
                  to ask mkdir to copy all the CHMOD_MODE_BITS, letting mkdir
                  decide what to do with S_ISUID | S_ISGID | S_ISVTX.  */
-              mkdir_mode = src_mode & CHMOD_MODE_BITS & ~omitted_permissions;
-              if (mkdir (dir, mkdir_mode) != 0)
+              mkdir_mode = x->explicit_no_preserve_mode ? S_IRWXUGO : src_mode;
+              mkdir_mode &= CHMOD_MODE_BITS & ~omitted_permissions;
+              if (mkdirat (dst_dirfd, dst_reldir, mkdir_mode) != 0)
                 {
                   error (0, errno, _("cannot make directory %s"),
-                         quote (dir));
+                         quoteaf (dir));
                   return false;
                 }
               else
@@ -489,10 +498,10 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
                  for writing the directory's contents. Check if these
                  permissions are there.  */
 
-              if (lstat (dir, &stats))
+              if (fstatat (dst_dirfd, dst_reldir, &stats, AT_SYMLINK_NOFOLLOW))
                 {
                   error (0, errno, _("failed to get attributes of %s"),
-                         quote (dir));
+                         quoteaf (dir));
                   return false;
                 }
 
@@ -509,15 +518,16 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
                     }
                 }
 
-              if ((stats.st_mode & S_IRWXU) != S_IRWXU)
+              mode_t accessible = stats.st_mode | S_IRWXU;
+              if (stats.st_mode != accessible)
                 {
                   /* Make the new directory searchable and writable.
                      The original permissions will be restored later.  */
 
-                  if (lchmod (dir, stats.st_mode | S_IRWXU) != 0)
+                  if (lchmodat (dst_dirfd, dst_reldir, accessible) != 0)
                     {
                       error (0, errno, _("setting permissions for %s"),
-                             quote (dir));
+                             quoteaf (dir));
                       return false;
                     }
                 }
@@ -525,14 +535,25 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
           else if (!S_ISDIR (stats.st_mode))
             {
               error (0, 0, _("%s exists but is not a directory"),
-                     quote (dir));
+                     quoteaf (dir));
               return false;
             }
           else
             *new_dst = false;
+
+          /* For existing dirs, set the security context as per that already
+             set for the process global context.  */
+          if (! *new_dst
+              && (x->set_security_context || x->preserve_security_context))
+            {
+              if (! set_file_security_ctx (dir, false, x)
+                  && x->require_preserve_context)
+                return false;
+            }
+
           *slash++ = '/';
 
-          /* Avoid unnecessary calls to `stat' when given
+          /* Avoid unnecessary calls to 'stat' when given
              file names containing multiple adjacent slashes.  */
           while (*slash == '/')
             slash++;
@@ -543,7 +564,7 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
 
   else if (!S_ISDIR (stats.st_mode))
     {
-      error (0, 0, _("%s exists but is not a directory"), quote (dst_dir));
+      error (0, 0, _("%s exists but is not a directory"), quoteaf (dst_dir));
       return false;
     }
   else
@@ -553,34 +574,11 @@ make_dir_parents_private (char const *const_dir, size_t src_offset,
   return true;
 }
 
-/* FILE is the last operand of this command.
-   Return true if FILE is a directory.
-   But report an error and exit if there is a problem accessing FILE,
-   or if FILE does not exist but would have to refer to an existing
-   directory if it referred to anything at all.
-
-   If the file exists, store the file's status into *ST.
-   Otherwise, set *NEW_DST.  */
-
-static bool
-target_directory_operand (char const *file, struct stat *st, bool *new_dst)
-{
-  int err = (stat (file, st) == 0 ? 0 : errno);
-  bool is_a_dir = !err && S_ISDIR (st->st_mode);
-  if (err)
-    {
-      if (err != ENOENT)
-        error (EXIT_FAILURE, err, _("accessing %s"), quote (file));
-      *new_dst = true;
-    }
-  return is_a_dir;
-}
-
 /* Scan the arguments, and copy each by calling copy.
    Return true if successful.  */
 
 static bool
-do_copy (int n_files, char **file, const char *target_directory,
+do_copy (int n_files, char **file, char const *target_directory,
          bool no_target_directory, struct cp_options *x)
 {
   struct stat sb;
@@ -593,40 +591,69 @@ do_copy (int n_files, char **file, const char *target_directory,
         error (0, 0, _("missing file operand"));
       else
         error (0, 0, _("missing destination file operand after %s"),
-               quote (file[0]));
+               quoteaf (file[0]));
       usage (EXIT_FAILURE);
     }
 
+  sb.st_mode = 0;
+  int target_dirfd = AT_FDCWD;
   if (no_target_directory)
     {
       if (target_directory)
-        error (EXIT_FAILURE, 0,
-               _("cannot combine --target-directory (-t) "
-                 "and --no-target-directory (-T)"));
+        die (EXIT_FAILURE, 0,
+             _("cannot combine --target-directory (-t) "
+               "and --no-target-directory (-T)"));
       if (2 < n_files)
         {
-          error (0, 0, _("extra operand %s"), quote (file[2]));
+          error (0, 0, _("extra operand %s"), quoteaf (file[2]));
           usage (EXIT_FAILURE);
         }
-      /* Update NEW_DST and SB, which may be checked below.  */
-      ignore_value (target_directory_operand (file[n_files -1], &sb, &new_dst));
     }
-  else if (!target_directory)
+  else if (target_directory)
     {
-      if (2 <= n_files
-          && target_directory_operand (file[n_files - 1], &sb, &new_dst))
-        target_directory = file[--n_files];
-      else if (2 < n_files)
-        error (EXIT_FAILURE, 0, _("target %s is not a directory"),
-               quote (file[n_files - 1]));
+      target_dirfd = target_directory_operand (target_directory, &sb);
+      if (! target_dirfd_valid (target_dirfd))
+        die (EXIT_FAILURE, errno, _("target directory %s"),
+             quoteaf (target_directory));
+    }
+  else
+    {
+      char const *lastfile = file[n_files - 1];
+      int fd = target_directory_operand (lastfile, &sb);
+      if (target_dirfd_valid (fd))
+        {
+          target_dirfd = fd;
+          target_directory = lastfile;
+          n_files--;
+        }
+      else
+        {
+          int err = errno;
+          if (err == ENOENT)
+            new_dst = true;
+
+          /* The last operand LASTFILE cannot be opened as a directory.
+             If there are more than two operands, report an error.
+
+             Also, report an error if LASTFILE is known to be a directory
+             even though it could not be opened, which can happen if
+             opening failed with EACCES on a platform lacking O_PATH.
+             In this case use stat to test whether LASTFILE is a
+             directory, in case opening a non-directory with (O_SEARCH
+             | O_DIRECTORY) failed with EACCES not ENOTDIR.  */
+          if (2 < n_files
+              || (O_PATHSEARCH == O_SEARCH && err == EACCES
+                  && (sb.st_mode || stat (lastfile, &sb) == 0)
+                  && S_ISDIR (sb.st_mode)))
+            die (EXIT_FAILURE, err, _("target %s"), quoteaf (lastfile));
+        }
     }
 
   if (target_directory)
     {
       /* cp file1...filen edir
-         Copy the files `file1' through `filen'
-         to the existing directory `edir'. */
-      int i;
+         Copy the files 'file1' through 'filen'
+         to the existing directory 'edir'. */
 
       /* Initialize these hash tables only if we'll need them.
          The problems they're used to detect can arise only if
@@ -637,7 +664,7 @@ do_copy (int n_files, char **file, const char *target_directory,
           src_info_init (x);
         }
 
-      for (i = 0; i < n_files; i++)
+      for (int i = 0; i < n_files; i++)
         {
           char *dst_name;
           bool parent_exists = true;  /* True if dir_name (dst_name) exists. */
@@ -654,15 +681,14 @@ do_copy (int n_files, char **file, const char *target_directory,
             {
               char *arg_no_trailing_slash;
 
-              /* Use `arg' without trailing slashes in constructing destination
+              /* Use 'arg' without trailing slashes in constructing destination
                  file names.  Otherwise, we can end up trying to create a
-                 directory via `mkdir ("dst/foo/"...', which is not portable.
-                 It fails, due to the trailing slash, on at least
+                 directory using a name with trailing slash, which fails on
                  NetBSD 1.[34] systems.  */
               ASSIGN_STRDUPA (arg_no_trailing_slash, arg);
               strip_trailing_slashes (arg_no_trailing_slash);
 
-              /* Append all of `arg' (minus any trailing slash) to `dest'.  */
+              /* Append all of 'arg' (minus any trailing slash) to 'dest'.  */
               dst_name = file_name_concat (target_directory,
                                            arg_no_trailing_slash,
                                            &arg_in_concat);
@@ -672,21 +698,23 @@ do_copy (int n_files, char **file, const char *target_directory,
                  leading directories. */
               parent_exists =
                 (make_dir_parents_private
-                 (dst_name, arg_in_concat - dst_name,
+                 (dst_name, arg_in_concat - dst_name, target_dirfd,
                   (x->verbose ? "%s -> %s\n" : NULL),
                   &attr_list, &new_dst, x));
+
+              while (*arg_in_concat == '/')
+                arg_in_concat++;
             }
           else
             {
               char *arg_base;
-              /* Append the last component of `arg' to `target_directory'.  */
-
-              ASSIGN_BASENAME_STRDUPA (arg_base, arg);
-              /* For `cp -R source/.. dest', don't copy into `dest/..'. */
-              dst_name = (STREQ (arg_base, "..")
-                          ? xstrdup (target_directory)
-                          : file_name_concat (target_directory, arg_base,
-                                              NULL));
+              /* Append the last component of 'arg' to 'target_directory'.  */
+              ASSIGN_STRDUPA (arg_base, last_component (arg));
+              strip_trailing_slashes (arg_base);
+              /* For 'cp -R source/.. dest', don't copy into 'dest/..'. */
+              arg_base += STREQ (arg_base, "..");
+              dst_name = file_name_concat (target_directory, arg_base,
+                                           &arg_in_concat);
             }
 
           if (!parent_exists)
@@ -698,10 +726,11 @@ do_copy (int n_files, char **file, const char *target_directory,
           else
             {
               bool copy_into_self;
-              ok &= copy (arg, dst_name, new_dst, x, &copy_into_self, NULL);
+              ok &= copy (arg, dst_name, target_dirfd, arg_in_concat,
+                          new_dst, x, &copy_into_self, NULL);
 
               if (parents_option)
-                ok &= re_protect (dst_name, arg_in_concat - dst_name,
+                ok &= re_protect (dst_name, target_dirfd, arg_in_concat,
                                   attr_list, x);
             }
 
@@ -720,7 +749,6 @@ do_copy (int n_files, char **file, const char *target_directory,
     }
   else /* !target_directory */
     {
-      char const *new_dest;
       char const *source = file[0];
       char const *dest = file[1];
       bool unused;
@@ -735,32 +763,29 @@ do_copy (int n_files, char **file, const char *target_directory,
       /* When the force and backup options have been specified and
          the source and destination are the same name for an existing
          regular file, convert the user's command, e.g.,
-         `cp --force --backup foo foo' to `cp --force foo fooSUFFIX'
+         'cp --force --backup foo foo' to 'cp --force foo fooSUFFIX'
          where SUFFIX is determined by any version control options used.  */
 
       if (x->unlink_dest_after_failed_open
           && x->backup_type != no_backups
           && STREQ (source, dest)
-          && !new_dst && S_ISREG (sb.st_mode))
+          && !new_dst
+          && (sb.st_mode != 0 || stat (dest, &sb) == 0) && S_ISREG (sb.st_mode))
         {
           static struct cp_options x_tmp;
 
-          new_dest = find_backup_file_name (dest, x->backup_type);
-          /* Set x->backup_type to `no_backups' so that the normal backup
+          dest = find_backup_file_name (AT_FDCWD, dest, x->backup_type);
+          /* Set x->backup_type to 'no_backups' so that the normal backup
              mechanism is not used when performing the actual copy.
-             backup_type must be set to `no_backups' only *after* the above
+             backup_type must be set to 'no_backups' only *after* the above
              call to find_backup_file_name -- that function uses
              backup_type to determine the suffix it applies.  */
           x_tmp = *x;
           x_tmp.backup_type = no_backups;
           x = &x_tmp;
         }
-      else
-        {
-          new_dest = dest;
-        }
 
-      ok = copy (source, new_dest, 0, x, &unused, NULL);
+      ok = copy (source, dest, AT_FDCWD, dest, -new_dst, x, &unused, NULL);
     }
 
   return ok;
@@ -777,15 +802,18 @@ cp_option_init (struct cp_options *x)
   x->hard_link = false;
   x->interactive = I_UNSPECIFIED;
   x->move_mode = false;
+  x->install_mode = false;
   x->one_file_system = false;
-  x->reflink_mode = REFLINK_NEVER;
+  x->reflink_mode = REFLINK_AUTO;
 
   x->preserve_ownership = false;
   x->preserve_links = false;
   x->preserve_mode = false;
   x->preserve_timestamps = false;
-  x->preserve_security_context = false;
-  x->require_preserve_context = false;
+  x->explicit_no_preserve_mode = false;
+  x->preserve_security_context = false; /* -a or --preserve=context.  */
+  x->require_preserve_context = false;  /* --preserve=context.  */
+  x->set_security_context = NULL;       /* -Z, set sys default context. */
   x->preserve_xattr = false;
   x->reduce_diagnostics = false;
   x->require_preserve_xattr = false;
@@ -835,8 +863,8 @@ decode_preserve_arg (char const *arg, struct cp_options *x, bool on_off)
       PRESERVE_OWNERSHIP, PRESERVE_LINK, PRESERVE_CONTEXT, PRESERVE_XATTR,
       PRESERVE_ALL
     };
-  /* Valid arguments to the `--preserve' option. */
-  static char const* const preserve_args[] =
+  /* Valid arguments to the '--preserve' option. */
+  static char const *const preserve_args[] =
     {
       "mode", "timestamps",
       "ownership", "links", "context", "xattr", "all", NULL
@@ -856,11 +884,13 @@ decode_preserve_arg (char const *arg, struct cp_options *x, bool on_off)
         *comma++ = 0;
 
       /* process S.  */
-      val = XARGMATCH ("--preserve", s, preserve_args, preserve_vals);
+      val = XARGMATCH (on_off ? "--preserve" : "--no-preserve",
+                       s, preserve_args, preserve_vals);
       switch (val)
         {
         case PRESERVE_MODE:
           x->preserve_mode = on_off;
+          x->explicit_no_preserve_mode = !on_off;
           break;
 
         case PRESERVE_TIMESTAMPS:
@@ -876,8 +906,8 @@ decode_preserve_arg (char const *arg, struct cp_options *x, bool on_off)
           break;
 
         case PRESERVE_CONTEXT:
-          x->preserve_security_context = on_off;
           x->require_preserve_context = on_off;
+          x->preserve_security_context = on_off;
           break;
 
         case PRESERVE_XATTR:
@@ -890,6 +920,7 @@ decode_preserve_arg (char const *arg, struct cp_options *x, bool on_off)
           x->preserve_timestamps = on_off;
           x->preserve_ownership = on_off;
           x->preserve_links = on_off;
+          x->explicit_no_preserve_mode = !on_off;
           if (selinux_enabled)
             x->preserve_security_context = on_off;
           x->preserve_xattr = on_off;
@@ -911,12 +942,13 @@ main (int argc, char **argv)
   int c;
   bool ok;
   bool make_backups = false;
-  char *backup_suffix_string;
+  char const *backup_suffix = NULL;
   char *version_control_string = NULL;
   struct cp_options x;
   bool copy_contents = false;
   char *target_directory = NULL;
   bool no_target_directory = false;
+  char const *scontext = NULL;
 
   initialize_main (&argc, &argv);
   set_program_name (argv[0]);
@@ -929,11 +961,7 @@ main (int argc, char **argv)
   selinux_enabled = (0 < is_selinux_enabled ());
   cp_option_init (&x);
 
-  /* FIXME: consider not calling getenv for SIMPLE_BACKUP_SUFFIX unless
-     we'll actually use backup_suffix_string.  */
-  backup_suffix_string = getenv ("SIMPLE_BACKUP_SUFFIX");
-
-  while ((c = getopt_long (argc, argv, "abdfHilLnprst:uvxPRS:T",
+  while ((c = getopt_long (argc, argv, "abdfHilLnprst:uvxPRS:TZ",
                            long_opts, NULL))
          != -1)
     {
@@ -977,6 +1005,10 @@ main (int argc, char **argv)
           x.data_copy_required = false;
           break;
 
+        case DEBUG_OPTION:
+          x.debug = x.verbose = true;
+          break;
+
         case COPY_CONTENTS_OPTION:
           copy_contents = true;
           break;
@@ -1021,7 +1053,7 @@ main (int argc, char **argv)
         case PRESERVE_ATTRIBUTES_OPTION:
           if (optarg == NULL)
             {
-              /* Fall through to the case for `p' below.  */
+              /* Fall through to the case for 'p' below.  */
             }
           else
             {
@@ -1029,6 +1061,7 @@ main (int argc, char **argv)
               x.require_preserve = true;
               break;
             }
+          FALLTHROUGH;
 
         case 'p':
           x.preserve_ownership = true;
@@ -1060,17 +1093,8 @@ main (int argc, char **argv)
 
         case 't':
           if (target_directory)
-            error (EXIT_FAILURE, 0,
-                   _("multiple target directories specified"));
-          else
-            {
-              struct stat st;
-              if (stat (optarg, &st) != 0)
-                error (EXIT_FAILURE, errno, _("accessing %s"), quote (optarg));
-              if (! S_ISDIR (st.st_mode))
-                error (EXIT_FAILURE, 0, _("target %s is not a directory"),
-                       quote (optarg));
-            }
+            die (EXIT_FAILURE, 0,
+                 _("multiple target directories specified"));
           target_directory = optarg;
           break;
 
@@ -1090,9 +1114,31 @@ main (int argc, char **argv)
           x.one_file_system = true;
           break;
 
+        case 'Z':
+          /* politely decline if we're not on a selinux-enabled kernel.  */
+          if (selinux_enabled)
+            {
+              if (optarg)
+                scontext = optarg;
+              else
+                {
+                  x.set_security_context = selabel_open (SELABEL_CTX_FILE,
+                                                         NULL, 0);
+                  if (! x.set_security_context)
+                    error (0, errno, _("warning: ignoring --context"));
+                }
+            }
+          else if (optarg)
+            {
+              error (0, 0,
+                     _("warning: ignoring --context; "
+                       "it requires an SELinux-enabled kernel"));
+            }
+          break;
+
         case 'S':
           make_backups = true;
-          backup_suffix_string = optarg;
+          backup_suffix = optarg;
           break;
 
         case_GETOPT_HELP_CHAR;
@@ -1110,6 +1156,9 @@ main (int argc, char **argv)
       usage (EXIT_FAILURE);
     }
 
+  if (x.interactive == I_ALWAYS_NO)
+    x.update = false;
+
   if (make_backups && x.interactive == I_ALWAYS_NO)
     {
       error (0, 0,
@@ -1123,17 +1172,15 @@ main (int argc, char **argv)
       usage (EXIT_FAILURE);
     }
 
-  if (backup_suffix_string)
-    simple_backup_suffix = xstrdup (backup_suffix_string);
-
   x.backup_type = (make_backups
                    ? xget_version (_("backup type"),
                                    version_control_string)
                    : no_backups);
+  set_simple_backup_suffix (backup_suffix);
 
   if (x.dereference == DEREF_UNDEFINED)
     {
-      if (x.recursive)
+      if (x.recursive && ! x.hard_link)
         /* This is compatible with FreeBSD.  */
         x.dereference = DEREF_NEVER;
       else
@@ -1143,23 +1190,35 @@ main (int argc, char **argv)
   if (x.recursive)
     x.copy_as_regular = copy_contents;
 
-  /* If --force (-f) was specified and we're in link-creation mode,
-     first remove any existing destination file.  */
-  if (x.unlink_dest_after_failed_open && (x.hard_link || x.symbolic_link))
-    x.unlink_dest_before_opening = true;
+  /* Ensure -Z overrides -a.  */
+  if ((x.set_security_context || scontext)
+      && ! x.require_preserve_context)
+    x.preserve_security_context = false;
 
-  if (x.preserve_security_context)
-    {
-      if (!selinux_enabled)
-        error (EXIT_FAILURE, 0,
-               _("cannot preserve security context "
-                 "without an SELinux-enabled kernel"));
-    }
+  if (x.preserve_security_context && (x.set_security_context || scontext))
+    die (EXIT_FAILURE, 0,
+         _("cannot set target context and preserve it"));
+
+  if (x.require_preserve_context && ! selinux_enabled)
+    die (EXIT_FAILURE, 0,
+         _("cannot preserve security context "
+           "without an SELinux-enabled kernel"));
+
+  /* FIXME: This handles new files.  But what about existing files?
+     I.e., if updating a tree, new files would have the specified context,
+     but shouldn't existing files be updated for consistency like this?
+       if (scontext && !restorecon (NULL, dst_path, 0))
+          error (...);
+   */
+  if (scontext && setfscreatecon (scontext) < 0)
+    die (EXIT_FAILURE, errno,
+         _("failed to set default file creation context to %s"),
+         quote (scontext));
 
 #if !USE_XATTR
   if (x.require_preserve_xattr)
-    error (EXIT_FAILURE, 0, _("cannot preserve extended attributes, cp is "
-                              "built without xattr support"));
+    die (EXIT_FAILURE, 0, _("cannot preserve extended attributes, cp is "
+                            "built without xattr support"));
 #endif
 
   /* Allocate space for remembering copied and created files.  */
@@ -1169,7 +1228,5 @@ main (int argc, char **argv)
   ok = do_copy (argc - optind, argv + optind,
                 target_directory, no_target_directory, &x);
 
-  forget_all ();
-
-  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
+  main_exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
 }
