@@ -1,5 +1,5 @@
 /* factor -- print prime factors of n.
-   Copyright (C) 1986-2021 Free Software Foundation, Inc.
+   Copyright (C) 1986-2023 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -226,12 +226,16 @@ enum
 
 static struct option const long_options[] =
 {
+  {"exponents", no_argument, NULL, 'h'},
   {"-debug", no_argument, NULL, DEV_DEBUG_OPTION},
   {GETOPT_HELP_OPTION_DECL},
   {GETOPT_VERSION_OPTION_DECL},
   {NULL, 0, NULL, 0}
 };
 
+/* If true, use p^e output format.  */
+static bool print_exponents;
+
 struct factors
 {
   uintmax_t     plarge[2]; /* Can have a single large factor */
@@ -431,7 +435,8 @@ mod2 (uintmax_t *r1, uintmax_t a1, uintmax_t a0, uintmax_t d1, uintmax_t d0)
   return a0;
 }
 
-static uintmax_t _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static uintmax_t
 gcd_odd (uintmax_t a, uintmax_t b)
 {
   if ((b & 1) == 0)
@@ -658,7 +663,7 @@ mp_factor_insert_ui (struct mp_factors *factors, unsigned long int prime)
 enum { W = sizeof (uintmax_t) * CHAR_BIT };
 
 /* Verify that uintmax_t does not have holes in its representation.  */
-verify (UINTMAX_MAX >> (W - 1) == 1);
+static_assert (UINTMAX_MAX >> (W - 1) == 1);
 
 #define P(a,b,c,d) a,
 static const unsigned char primes_diff[] = {
@@ -691,7 +696,7 @@ static const struct primes_dtab primes_dtab[] = {
 
 /* Verify that uintmax_t is not wider than
    the integers used to generate primes.h.  */
-verify (W <= WIDE_UINT_BITS);
+static_assert (W <= WIDE_UINT_BITS);
 
 /* debugging for developers.  Enables devmsg().
    This flag is used only in the GMP code.  */
@@ -775,7 +780,8 @@ factor_using_division (uintmax_t *t1p, uintmax_t t1, uintmax_t t0,
     {
       for (;;)
         {
-          uintmax_t q1, q0, hi, lo _GL_UNUSED;
+          uintmax_t q1, q0, hi;
+          MAYBE_UNUSED uintmax_t lo;
 
           q0 = t0 * primes_dtab[i].binv;
           umul_ppmm (hi, lo, q0, p);
@@ -915,7 +921,8 @@ static const unsigned char  binvert_table[128] =
     _q0 = (u0) * _di;                                                   \
     if ((u1) >= (d))                                                    \
       {                                                                 \
-        uintmax_t _p1, _p0 _GL_UNUSED;					\
+        uintmax_t _p1;                                                  \
+        MAYBE_UNUSED intmax_t _p0;                                      \
         umul_ppmm (_p1, _p0, _q0, d);                                   \
         (q1) = ((u1) - _p1) * _di;                                      \
         (q0) = _q0;                                                     \
@@ -930,7 +937,7 @@ static const unsigned char  binvert_table[128] =
 /* x B (mod n).  */
 #define redcify(r_prim, r, n)                                           \
   do {                                                                  \
-    uintmax_t _redcify_q _GL_UNUSED;					\
+    MAYBE_UNUSED uintmax_t _redcify_q;					\
     udiv_qrnnd (_redcify_q, r_prim, r, 0, n);                           \
   } while (0)
 
@@ -946,7 +953,7 @@ static const unsigned char  binvert_table[128] =
     else                                                                \
       {                                                                 \
         _r1 = 0; _r0 = (x);                                             \
-        _i = 2*W_TYPE_SIZE;                                             \
+        _i = 2 * W_TYPE_SIZE;                                           \
       }                                                                 \
     while (_i-- > 0)                                                    \
       {                                                                 \
@@ -963,7 +970,8 @@ static const unsigned char  binvert_table[128] =
 static inline uintmax_t
 mulredc (uintmax_t a, uintmax_t b, uintmax_t m, uintmax_t mi)
 {
-  uintmax_t rh, rl, q, th, tl _GL_UNUSED, xh;
+  uintmax_t rh, rl, q, th, xh;
+  MAYBE_UNUSED uintmax_t tl;
 
   umul_ppmm (rh, rl, a, b);
   q = rl * mi;
@@ -983,7 +991,8 @@ mulredc2 (uintmax_t *r1p,
           uintmax_t a1, uintmax_t a0, uintmax_t b1, uintmax_t b0,
           uintmax_t m1, uintmax_t m0, uintmax_t mi)
 {
-  uintmax_t r1, r0, q, p1, p0 _GL_UNUSED, t1, t0, s1, s0;
+  uintmax_t r1, r0, q, p1, t1, t0, s1, s0;
+  MAYBE_UNUSED uintmax_t p0;
   mi = -mi;
   assert ((a1 >> (W_TYPE_SIZE - 1)) == 0);
   assert ((b1 >> (W_TYPE_SIZE - 1)) == 0);
@@ -1046,7 +1055,8 @@ mulredc2 (uintmax_t *r1p,
   return r0;
 }
 
-static uintmax_t _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static uintmax_t
 powm (uintmax_t b, uintmax_t e, uintmax_t n, uintmax_t ni, uintmax_t one)
 {
   uintmax_t y = one;
@@ -1107,7 +1117,8 @@ powm2 (uintmax_t *r1m,
   return r0;
 }
 
-static bool _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static bool
 millerrabin (uintmax_t n, uintmax_t ni, uintmax_t b, uintmax_t q,
              unsigned int k, uintmax_t one)
 {
@@ -1130,7 +1141,7 @@ millerrabin (uintmax_t n, uintmax_t ni, uintmax_t b, uintmax_t q,
   return false;
 }
 
-static bool
+ATTRIBUTE_PURE static bool
 millerrabin2 (const uintmax_t *np, uintmax_t ni, const uintmax_t *bp,
               const uintmax_t *qp, unsigned int k, const uintmax_t *one)
 {
@@ -1251,7 +1262,7 @@ prime_p (uintmax_t n)
           a_prim = s0 % n;
         else
           {
-            uintmax_t dummy _GL_UNUSED;
+            MAYBE_UNUSED uintmax_t dummy;
             udiv_qrnnd (dummy, a_prim, s1, s0, n);
           }
       }
@@ -1752,7 +1763,8 @@ mp_factor_using_pollard_rho (mpz_t n, unsigned long int a,
 #if USE_SQUFOF
 /* FIXME: Maybe better to use an iteration converging to 1/sqrt(n)?  If
    algorithm is replaced, consider also returning the remainder.  */
-static uintmax_t _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static uintmax_t
 isqrt (uintmax_t n)
 {
   uintmax_t x;
@@ -1767,7 +1779,7 @@ isqrt (uintmax_t n)
 
   for (;;)
     {
-      uintmax_t y = (x + n/x) / 2;
+      uintmax_t y = (x + n / x) / 2;
       if (y >= x)
         return x;
 
@@ -1775,7 +1787,8 @@ isqrt (uintmax_t n)
     }
 }
 
-static uintmax_t _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static uintmax_t
 isqrt2 (uintmax_t nh, uintmax_t nl)
 {
   unsigned int shift;
@@ -1797,7 +1810,7 @@ isqrt2 (uintmax_t nh, uintmax_t nl)
   /* Do we need more than one iteration?  */
   for (;;)
     {
-      uintmax_t r _GL_UNUSED;
+      MAYBE_UNUSED uintmax_t r;
       uintmax_t q, y;
       udiv_qrnnd (q, r, nh, nl, x);
       y = (x + q) / 2;
@@ -1827,7 +1840,8 @@ isqrt2 (uintmax_t nh, uintmax_t nl)
 # define MAGIC11 0x23b
 
 /* Return the square root if the input is a square, otherwise 0.  */
-static uintmax_t _GL_ATTRIBUTE_CONST
+ATTRIBUTE_CONST
+static uintmax_t
 is_square (uintmax_t x)
 {
   /* Uses the tests suggested by Cohen.  Excludes 99% of the non-squares before
@@ -1839,7 +1853,7 @@ is_square (uintmax_t x)
       && ((MAGIC11 >> (x % 11) & 1)))
     {
       uintmax_t r = isqrt (x);
-      if (r*r == x)
+      if (r * r == x)
         return r;
     }
   return 0;
@@ -1887,7 +1901,7 @@ static const unsigned short invtab[0x81] =
             _dinv = invtab[((d) >> (W_TYPE_SIZE - 8 - _cnt)) - 0x7f];   \
             _q = _dinv * (_r >> (W_TYPE_SIZE - 3 - _cnt)) >> 11;        \
           }                                                             \
-        _r -= _q*(d);                                                   \
+        _r -= _q * (d);                                                 \
                                                                         \
         _mask = -(uintmax_t) (_r >= (d));                               \
         (r) = _r - (_mask & (d));                                       \
@@ -1903,11 +1917,11 @@ static const unsigned short invtab[0x81] =
   } while (0)
 
 /* Notes: Example N = 22117019.  After first phase we find Q1 = 6314, Q
-   = 3025, P = 1737, representing F_{18} = (-6314, 2* 1737, 3025),
+   = 3025, P = 1737, representing F_{18} = (-6314, 2 * 1737, 3025),
    with 3025 = 55^2.
 
    Constructing the square root, we get Q1 = 55, Q = 8653, P = 4652,
-   representing G_0 = (-55, 2*4652, 8653).
+   representing G_0 = (-55, 2 * 4652, 8653).
 
    In the notation of the paper:
 
@@ -1944,7 +1958,6 @@ static const unsigned short invtab[0x81] =
 # define Q_FREQ_SIZE 50
 /* Element 0 keeps the total */
 static unsigned int q_freq[Q_FREQ_SIZE + 1];
-# define MIN(a,b) ((a) < (b) ? (a) : (b))
 #endif
 
 #if USE_SQUFOF
@@ -2002,7 +2015,7 @@ factor_using_squfof (uintmax_t n1, uintmax_t n0, struct factors *factors)
                 }
               /* Duplicate the new factors */
               for (unsigned int i = 0; i < f.nfactors; i++)
-                factor_insert_multiplicity (factors, f.p[i], 2*f.e[i]);
+                factor_insert_multiplicity (factors, f.p[i], 2 * f.e[i]);
             }
           return true;
         }
@@ -2031,7 +2044,7 @@ factor_using_squfof (uintmax_t n1, uintmax_t n0, struct factors *factors)
          Requiring 64 mu^3 < n seems sufficient.  */
       if (n1 == 0)
         {
-          if ((uintmax_t) mu*mu*mu >= n0 / 64)
+          if ((uintmax_t) mu * mu * mu >= n0 / 64)
             continue;
         }
       else
@@ -2051,10 +2064,10 @@ factor_using_squfof (uintmax_t n1, uintmax_t n0, struct factors *factors)
       P = S;
 
       /* Square root remainder fits in one word, so ignore high part.  */
-      Q = Dl - P*P;
+      Q = Dl - P * P;
       /* FIXME: When can this differ from floor (sqrt (2 * sqrt (D)))?  */
-      L = isqrt (2*S);
-      B = 2*L;
+      L = isqrt (2 * S);
+      B = 2 * L;
       L1 = mu * 2 * L;
 
       /* The form is (+/- Q1, 2P, -/+ Q), of discriminant 4 (P^2 + Q Q1) =
@@ -2064,10 +2077,10 @@ factor_using_squfof (uintmax_t n1, uintmax_t n0, struct factors *factors)
         {
           uintmax_t q, P1, t, rem;
 
-          div_smallq (q, rem, S+P, Q);
+          div_smallq (q, rem, S + P, Q);
           P1 = S - rem; /* P1 = q*Q - P */
 
-          IF_LINT (assert (q > 0 && Q > 0));
+          assert (q > 0 && Q > 0);
 
 # if STAT_SQUFOF
           q_freq[0]++;
@@ -2153,7 +2166,7 @@ factor_using_squfof (uintmax_t n1, uintmax_t n0, struct factors *factors)
                          floor([S+P]/\hat Q), but looking at the equations
                          in Sec. 3.1, it should be q <-- floor([S+P] / Q).
                          (In this code, \hat Q is Q1).  */
-                      div_smallq (q, rem, S+P, Q);
+                      div_smallq (q, rem, S + P, Q);
                       P1 = S - rem;     /* P1 = q*Q - P */
 
 # if STAT_SQUFOF
@@ -2448,6 +2461,12 @@ print_factors_single (uintmax_t t1, uintmax_t t0)
       {
         lbuf_putc (' ');
         print_uintmaxes (0, factors.p[j]);
+        if (print_exponents && factors.e[j] > 1)
+          {
+            lbuf_putc ('^');
+            lbuf_putint (factors.e[j], 0);
+            break;
+          }
       }
 
   if (factors.plarge[1])
@@ -2516,6 +2535,11 @@ print_factors (char const *input)
       {
         putchar (' ');
         mpz_out_str (stdout, 10, factors.p[j]);
+        if (print_exponents && factors.e[j] > 1)
+          {
+            printf ("^%lu", factors.e[j]);
+            break;
+          }
       }
 
   mp_factor_clear (&factors);
@@ -2533,15 +2557,17 @@ usage (int status)
   else
     {
       printf (_("\
-Usage: %s [NUMBER]...\n\
-  or:  %s OPTION\n\
+Usage: %s [OPTION] [NUMBER]...\n\
 "),
-              program_name, program_name);
+              program_name);
       fputs (_("\
 Print the prime factors of each specified integer NUMBER.  If none\n\
 are specified on the command line, read them from standard input.\n\
 \n\
 "), stdout);
+      fputs ("\
+  -h, --exponents   print repeated factors in form p^e unless e is 1\n\
+", stdout);
       fputs (HELP_OPTION_DESCRIPTION, stdout);
       fputs (VERSION_OPTION_DESCRIPTION, stdout);
       emit_ancillary_info (PROGRAM_NAME);
@@ -2584,10 +2610,14 @@ main (int argc, char **argv)
   atexit (lbuf_flush);
 
   int c;
-  while ((c = getopt_long (argc, argv, "", long_options, NULL)) != -1)
+  while ((c = getopt_long (argc, argv, "h", long_options, NULL)) != -1)
     {
       switch (c)
         {
+        case 'h':  /* NetBSD used -h for this functionality first.  */
+          print_exponents = true;
+          break;
+
         case DEV_DEBUG_OPTION:
           dev_debug = true;
           break;
