@@ -1,5 +1,5 @@
 /* od -- dump files in octal and other formats
-   Copyright (C) 1992-2014 Free Software Foundation, Inc.
+   Copyright (C) 1992-2023 Free Software Foundation, Inc.
 
    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -12,7 +12,7 @@
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
-   along with this program.  If not, see <http://www.gnu.org/licenses/>.  */
+   along with this program.  If not, see <https://www.gnu.org/licenses/>.  */
 
 /* Written by Jim Meyering.  */
 
@@ -24,12 +24,15 @@
 #include <sys/types.h>
 #include "system.h"
 #include "argmatch.h"
+#include "die.h"
 #include "error.h"
 #include "ftoastr.h"
 #include "quote.h"
-#include "xfreopen.h"
+#include "stat-size.h"
+#include "xbinary-io.h"
 #include "xprintf.h"
 #include "xstrtol.h"
+#include "xstrtol-error.h"
 
 /* The official name of this program (e.g., no 'g' prefix).  */
 #define PROGRAM_NAME "od"
@@ -91,7 +94,7 @@ enum
   };
 
 /* Ensure that our choice for FMT_BYTES_ALLOCATED is reasonable.  */
-verify (MAX_INTEGRAL_TYPE_SIZE * CHAR_BIT / 3 <= 99);
+static_assert (MAX_INTEGRAL_TYPE_SIZE * CHAR_BIT / 3 <= 99);
 
 /* Each output format specification (from '-t spec' or from
    old-style options) is represented by one of these structures.  */
@@ -136,12 +139,14 @@ static unsigned int const bytes_to_hex_digits[] =
 /* It'll be a while before we see integral types wider than 16 bytes,
    but if/when it happens, this check will catch it.  Without this check,
    a wider type would provoke a buffer overrun.  */
-verify (MAX_INTEGRAL_TYPE_SIZE < ARRAY_CARDINALITY (bytes_to_hex_digits));
+static_assert (MAX_INTEGRAL_TYPE_SIZE
+               < ARRAY_CARDINALITY (bytes_to_hex_digits));
 
 /* Make sure the other arrays have the same length.  */
-verify (sizeof bytes_to_oct_digits == sizeof bytes_to_signed_dec_digits);
-verify (sizeof bytes_to_oct_digits == sizeof bytes_to_unsigned_dec_digits);
-verify (sizeof bytes_to_oct_digits == sizeof bytes_to_hex_digits);
+static_assert (sizeof bytes_to_oct_digits == sizeof bytes_to_signed_dec_digits);
+static_assert (sizeof bytes_to_oct_digits
+               == sizeof bytes_to_unsigned_dec_digits);
+static_assert (sizeof bytes_to_oct_digits == sizeof bytes_to_hex_digits);
 
 /* Convert enum size_spec to the size of the named type.  */
 static const int width_bytes[] =
@@ -159,7 +164,7 @@ static const int width_bytes[] =
 
 /* Ensure that for each member of 'enum size_spec' there is an
    initializer in the width_bytes array.  */
-verify (ARRAY_CARDINALITY (width_bytes) == N_SIZE_SPECS);
+static_assert (ARRAY_CARDINALITY (width_bytes) == N_SIZE_SPECS);
 
 /* Names for some non-printing characters.  */
 static char const charname[33][4] =
@@ -327,10 +332,12 @@ Usage: %s [OPTION]... [FILE]...\n\
 Write an unambiguous representation, octal bytes by default,\n\
 of FILE to standard output.  With more than one FILE argument,\n\
 concatenate them in the listed order to form the input.\n\
-With no FILE, or when FILE is -, read standard input.\n\
-\n\
 "), stdout);
+
+      emit_stdin_note ();
+
       fputs (_("\
+\n\
 If first and second call formats both apply, the second format is assumed\n\
 if the last operand begins with + or (if there are 2 operands) a digit.\n\
 An OFFSET operand means -j OFFSET.  LABEL is the pseudo-address\n\
@@ -386,7 +393,7 @@ TYPE is made up of one or more of these specifications:\n\
 "), stdout);
       fputs (_("\
   d[SIZE]    signed decimal, SIZE bytes per integer\n\
-  f[SIZE]    floating point, SIZE bytes per integer\n\
+  f[SIZE]    floating point, SIZE bytes per float\n\
   o[SIZE]    octal, SIZE bytes per integer\n\
   u[SIZE]    unsigned decimal, SIZE bytes per integer\n\
   x[SIZE]    hexadecimal, SIZE bytes per integer\n\
@@ -412,19 +419,20 @@ BYTES is hex with 0x or 0X prefix, and may have a multiplier suffix:\n\
   K    1024\n\
   MB   1000*1000\n\
   M    1024*1024\n\
-and so on for G, T, P, E, Z, Y.\n\
+and so on for G, T, P, E, Z, Y, R, Q.\n\
+Binary prefixes can be used, too: KiB=K, MiB=M, and so on.\n\
 "), stdout);
-      emit_ancillary_info ();
+      emit_ancillary_info (PROGRAM_NAME);
     }
   exit (status);
 }
 
 /* Define the print functions.  */
 
-#define PRINT_FIELDS(N, T, FMT_STRING, ACTION)                          \
+#define PRINT_FIELDS(N, T, FMT_STRING_DECL, ACTION)                     \
 static void                                                             \
 N (size_t fields, size_t blank, void const *block,                      \
-   char const *FMT_STRING, int width, int pad)                          \
+   FMT_STRING_DECL, int width, int pad)                                 \
 {                                                                       \
   T const *p = block;                                                   \
   uintmax_t i;                                                          \
@@ -442,7 +450,7 @@ N (size_t fields, size_t blank, void const *block,                      \
             char b[sizeof (T)];                                         \
           } u;                                                          \
           for (j = 0; j < sizeof (T); j++)                              \
-            u.b[j] = ((const char *) p)[sizeof (T) - 1 - j];            \
+            u.b[j] = ((char const *) p)[sizeof (T) - 1 - j];            \
           x = u.x;                                                      \
         }                                                               \
       else                                                              \
@@ -454,10 +462,11 @@ N (size_t fields, size_t blank, void const *block,                      \
 }
 
 #define PRINT_TYPE(N, T)                                                \
-  PRINT_FIELDS (N, T, fmt_string, xprintf (fmt_string, adjusted_width, x))
+  PRINT_FIELDS (N, T, char const *fmt_string,                           \
+                xprintf (fmt_string, adjusted_width, x))
 
 #define PRINT_FLOATTYPE(N, T, FTOASTR, BUFSIZE)                         \
-  PRINT_FIELDS (N, T, fmt_string _GL_UNUSED,                      \
+  PRINT_FIELDS (N, T, MAYBE_UNUSED char const *fmt_string,              \
                 char buf[BUFSIZE];                                      \
                 FTOASTR (buf, sizeof buf, 0, 0, x);                     \
                 xprintf ("%*s", adjusted_width, buf))
@@ -478,11 +487,10 @@ PRINT_FLOATTYPE (print_long_double, long double, ldtoastr, LDBL_BUFSIZE_BOUND)
 #undef PRINT_FLOATTYPE
 
 static void
-dump_hexl_mode_trailer (size_t n_bytes, const char *block)
+dump_hexl_mode_trailer (size_t n_bytes, char const *block)
 {
-  size_t i;
   fputs ("  >", stdout);
-  for (i = n_bytes; i > 0; i--)
+  for (size_t i = n_bytes; i > 0; i--)
     {
       unsigned char c = *block++;
       unsigned char c2 = (isprint (c) ? c : '.');
@@ -493,7 +501,7 @@ dump_hexl_mode_trailer (size_t n_bytes, const char *block)
 
 static void
 print_named_ascii (size_t fields, size_t blank, void const *block,
-                   const char *unused_fmt_string _GL_UNUSED,
+                   MAYBE_UNUSED char const *unused_fmt_string,
                    int width, int pad)
 {
   unsigned char const *p = block;
@@ -503,7 +511,7 @@ print_named_ascii (size_t fields, size_t blank, void const *block,
     {
       int next_pad = pad * (i - 1) / fields;
       int masked_c = *p++ & 0x7f;
-      const char *s;
+      char const *s;
       char buf[2];
 
       if (masked_c == 127)
@@ -524,7 +532,7 @@ print_named_ascii (size_t fields, size_t blank, void const *block,
 
 static void
 print_ascii (size_t fields, size_t blank, void const *block,
-             const char *unused_fmt_string _GL_UNUSED, int width,
+             MAYBE_UNUSED char const *unused_fmt_string, int width,
              int pad)
 {
   unsigned char const *p = block;
@@ -534,7 +542,7 @@ print_ascii (size_t fields, size_t blank, void const *block,
     {
       int next_pad = pad * (i - 1) / fields;
       unsigned char c = *p++;
-      const char *s;
+      char const *s;
       char buf[4];
 
       switch (c)
@@ -590,7 +598,7 @@ print_ascii (size_t fields, size_t blank, void const *block,
    the result of the conversion and return true.  */
 
 static bool
-simple_strtoul (const char *s, const char **p, unsigned long int *val)
+simple_strtoul (char const *s, char const **p, unsigned long int *val)
 {
   unsigned long int sum;
 
@@ -627,7 +635,7 @@ simple_strtoul (const char *s, const char **p, unsigned long int *val)
    */
 
 static bool
-decode_one_format (const char *s_orig, const char *s, const char **next,
+decode_one_format (char const *s_orig, char const *s, char const **next,
                    struct tspec *tspec)
 {
   enum size_spec size_spec;
@@ -635,7 +643,7 @@ decode_one_format (const char *s_orig, const char *s, const char **next,
   enum output_format fmt;
   void (*print_function) (size_t, size_t, void const *, char const *,
                           int, int);
-  const char *p;
+  char const *p;
   char c;
   int field_width;
 
@@ -910,15 +918,14 @@ open_next_file (void)
           input_filename = _("standard input");
           in_stream = stdin;
           have_read_stdin = true;
-          if (O_BINARY && ! isatty (STDIN_FILENO))
-            xfreopen (NULL, "rb", stdin);
+          xset_binary_mode (STDIN_FILENO, O_BINARY);
         }
       else
         {
           in_stream = fopen (input_filename, (O_BINARY ? "rb" : "r"));
           if (in_stream == NULL)
             {
-              error (0, errno, "%s", input_filename);
+              error (0, errno, "%s", quotef (input_filename));
               ok = false;
             }
         }
@@ -945,16 +952,15 @@ check_and_close (int in_errno)
 
   if (in_stream != NULL)
     {
-      if (ferror (in_stream))
+      if (!ferror (in_stream))
+        in_errno = 0;
+      if (STREQ (file_list[-1], "-"))
+        clearerr (in_stream);
+      else if (fclose (in_stream) != 0 && !in_errno)
+        in_errno = errno;
+      if (in_errno)
         {
-          error (0, in_errno, _("%s: read error"), input_filename);
-          if (! STREQ (file_list[-1], "-"))
-            fclose (in_stream);
-          ok = false;
-        }
-      else if (! STREQ (file_list[-1], "-") && fclose (in_stream) != 0)
-        {
-          error (0, errno, "%s", input_filename);
+          error (0, in_errno, "%s", quotef (input_filename));
           ok = false;
         }
 
@@ -975,14 +981,14 @@ check_and_close (int in_errno)
    necessary.  Return true if S is valid.  */
 
 static bool
-decode_format_string (const char *s)
+decode_format_string (char const *s)
 {
-  const char *s_orig = s;
+  char const *s_orig = s;
   assert (s != NULL);
 
   while (*s != '\0')
     {
-      const char *next;
+      char const *next;
 
       if (n_specs_allocated <= n_specs)
         spec = X2NREALLOC (spec, &n_specs_allocated);
@@ -1030,13 +1036,16 @@ skip (uintmax_t n_skip)
 
       if (fstat (fileno (in_stream), &file_stats) == 0)
         {
+          bool usable_size = usable_st_size (&file_stats);
+
           /* The st_size field is valid for regular files.
              If the number of bytes left to skip is larger than
              the size of the current file, we can decrement n_skip
              and go on to the next file.  Skip this optimization also
-             when st_size is 0, because some kernels report that
-             nonempty files in /proc have st_size == 0.  */
-          if (S_ISREG (file_stats.st_mode) && 0 < file_stats.st_size)
+             when st_size is no greater than the block size, because
+             some kernels report nonsense small file sizes for
+             proc-like file systems.  */
+          if (usable_size && ST_BLKSIZE (file_stats) < file_stats.st_size)
             {
               if ((uintmax_t) file_stats.st_size < n_skip)
                 n_skip -= file_stats.st_size;
@@ -1051,7 +1060,11 @@ skip (uintmax_t n_skip)
                 }
             }
 
+          else if (!usable_size && fseeko (in_stream, n_skip, SEEK_CUR) == 0)
+            n_skip = 0;
+
           /* If it's not a regular file with nonnegative size,
+             or if it's so small that it might be in a proc-like file system,
              position the file pointer by reading.  */
 
           else
@@ -1067,10 +1080,15 @@ skip (uintmax_t n_skip)
                   n_skip -= n_bytes_read;
                   if (n_bytes_read != n_bytes_to_read)
                     {
-                      in_errno = errno;
-                      ok = false;
-                      n_skip = 0;
-                      break;
+                      if (ferror (in_stream))
+                        {
+                          in_errno = errno;
+                          ok = false;
+                          n_skip = 0;
+                          break;
+                        }
+                      if (feof (in_stream))
+                        break;
                     }
                 }
             }
@@ -1081,7 +1099,7 @@ skip (uintmax_t n_skip)
 
       else   /* cannot fstat() file */
         {
-          error (0, errno, "%s", input_filename);
+          error (0, errno, "%s", quotef (input_filename));
           ok = false;
         }
 
@@ -1091,14 +1109,14 @@ skip (uintmax_t n_skip)
     }
 
   if (n_skip != 0)
-    error (EXIT_FAILURE, 0, _("cannot skip past end of combined input"));
+    die (EXIT_FAILURE, 0, _("cannot skip past end of combined input"));
 
   return ok;
 }
 
 static void
-format_address_none (uintmax_t address _GL_UNUSED,
-                     char c _GL_UNUSED)
+format_address_none (MAYBE_UNUSED uintmax_t address,
+                     MAYBE_UNUSED char c)
 {
 }
 
@@ -1170,7 +1188,7 @@ format_address_label (uintmax_t address, char c)
 
 static void
 write_block (uintmax_t current_offset, size_t n_bytes,
-             const char *prev_block, const char *curr_block)
+             char const *prev_block, char const *curr_block)
 {
   static bool first = true;
   static bool prev_pair_equal = false;
@@ -1194,10 +1212,8 @@ write_block (uintmax_t current_offset, size_t n_bytes,
     }
   else
     {
-      size_t i;
-
       prev_pair_equal = false;
-      for (i = 0; i < n_specs; i++)
+      for (size_t i = 0; i < n_specs; i++)
         {
           int datum_width = width_bytes[spec[i].size];
           int fields_per_block = bytes_per_block / datum_width;
@@ -1280,9 +1296,6 @@ read_block (size_t n, char *block, size_t *n_bytes_in_buffer)
 
   *n_bytes_in_buffer = 0;
 
-  if (n == 0)
-    return true;
-
   while (in_stream != NULL)	/* EOF.  */
     {
       size_t n_needed;
@@ -1307,13 +1320,13 @@ read_block (size_t n, char *block, size_t *n_bytes_in_buffer)
 /* Return the least common multiple of the sizes associated
    with the format specs.  */
 
-static int _GL_ATTRIBUTE_PURE
+ATTRIBUTE_PURE
+static int
 get_lcm (void)
 {
-  size_t i;
   int l_c_m = 1;
 
-  for (i = 0; i < n_specs; i++)
+  for (size_t i = 0; i < n_specs; i++)
     l_c_m = lcm (l_c_m, width_bytes[spec[i].size]);
   return l_c_m;
 }
@@ -1322,7 +1335,7 @@ get_lcm (void)
    leading '+' return true and set *OFFSET to the offset it denotes.  */
 
 static bool
-parse_old_offset (const char *s, uintmax_t *offset)
+parse_old_offset (char const *s, uintmax_t *offset)
 {
   int radix;
 
@@ -1377,7 +1390,7 @@ dump (void)
 
   if (limit_bytes_to_format)
     {
-      while (1)
+      while (true)
         {
           size_t n_needed;
           if (current_offset >= end_offset)
@@ -1399,7 +1412,7 @@ dump (void)
     }
   else
     {
-      while (1)
+      while (true)
         {
           ok &= read_block (bytes_per_block, block[idx], &n_bytes_read);
           if (n_bytes_read < bytes_per_block)
@@ -1452,7 +1465,7 @@ dump_strings (void)
   uintmax_t address = n_bytes_to_skip;
   bool ok = true;
 
-  while (1)
+  while (true)
     {
       size_t i;
       int c;
@@ -1565,7 +1578,7 @@ main (int argc, char **argv)
   bool width_specified = false;
   bool ok = true;
   size_t width_per_block = 0;
-  static char const multipliers[] = "bEGKkMmPTYZ0";
+  static char const multipliers[] = "bEGKkMmPQRTYZ0";
 
   /* The old-style 'pseudo starting address' to be printed in parentheses
      after any true address.  */
@@ -1646,10 +1659,10 @@ main (int argc, char **argv)
               address_pad_len = 0;
               break;
             default:
-              error (EXIT_FAILURE, 0,
-                     _("invalid output address radix '%c';\
+              die (EXIT_FAILURE, 0,
+                   _("invalid output address radix '%c';\
  it must be one character from [doxn]"),
-                     optarg[0]);
+                   optarg[0]);
               break;
             }
           break;
@@ -1684,7 +1697,7 @@ main (int argc, char **argv)
               /* The minimum string length may be no larger than SIZE_MAX,
                  since we may allocate a buffer of this size.  */
               if (SIZE_MAX < tmp)
-                error (EXIT_FAILURE, 0, _("%s is too large"), optarg);
+                die (EXIT_FAILURE, 0, _("%s is too large"), quote (optarg));
 
               string_min = tmp;
             }
@@ -1765,7 +1778,7 @@ main (int argc, char **argv)
               if (s_err != LONGINT_OK)
                 xstrtol_fatal (s_err, oi, c, long_options, optarg);
               if (SIZE_MAX < w_tmp)
-                error (EXIT_FAILURE, 0, _("%s is too large"), optarg);
+                die (EXIT_FAILURE, 0, _("%s is too large"), quote (optarg));
               desired_width = w_tmp;
             }
           break;
@@ -1781,11 +1794,11 @@ main (int argc, char **argv)
     }
 
   if (!ok)
-    exit (EXIT_FAILURE);
+    return EXIT_FAILURE;
 
   if (flag_dump_strings && n_specs > 0)
-    error (EXIT_FAILURE, 0,
-           _("no type may be specified when dumping strings"));
+    die (EXIT_FAILURE, 0,
+         _("no type may be specified when dumping strings"));
 
   n_files = argc - optind;
 
@@ -1881,7 +1894,7 @@ main (int argc, char **argv)
     {
       end_offset = n_bytes_to_skip + max_bytes_to_format;
       if (end_offset < n_bytes_to_skip)
-        error (EXIT_FAILURE, 0, _("skip-bytes + read-bytes is too large"));
+        die (EXIT_FAILURE, 0, _("skip-bytes + read-bytes is too large"));
     }
 
   if (n_specs == 0)
@@ -1953,7 +1966,8 @@ main (int argc, char **argv)
     }
 
 #ifdef DEBUG
-  printf ("lcm=%d, width_per_block=%zu\n", l_c_m, width_per_block);
+  printf ("lcm=%d, width_per_block=%"PRIuMAX"\n", l_c_m,
+          (uintmax_t) width_per_block);
   for (i = 0; i < n_specs; i++)
     {
       int fields_per_block = bytes_per_block / width_bytes[spec[i].size];
@@ -1970,7 +1984,7 @@ main (int argc, char **argv)
 cleanup:
 
   if (have_read_stdin && fclose (stdin) == EOF)
-    error (EXIT_FAILURE, errno, _("standard input"));
+    die (EXIT_FAILURE, errno, _("standard input"));
 
-  exit (ok ? EXIT_SUCCESS : EXIT_FAILURE);
+  return ok ? EXIT_SUCCESS : EXIT_FAILURE;
 }
