@@ -1,7 +1,7 @@
 #!/bin/sh
 # Validate timeout basic operation
 
-# Copyright (C) 2008-2013 Free Software Foundation, Inc.
+# Copyright (C) 2008-2023 Free Software Foundation, Inc.
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -14,10 +14,11 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 . "${srcdir=.}/tests/init.sh"; path_prepend_ ./src
 print_ver_ timeout
+require_trap_signame_
 
 # no timeout
 timeout 10 true || fail=1
@@ -29,29 +30,27 @@ timeout 1d true || fail=1
 timeout 0 true || fail=1
 
 # exit status propagation
-timeout 10 sh -c 'exit 2'
-test $? = 2 || fail=1
+returns_ 2 timeout 10 sh -c 'exit 2' || fail=1
 
 # timeout
-timeout .1 sleep 10
-test $? = 124 || fail=1
+returns_ 124 timeout .1 sleep 10 || fail=1
 
 # exit status propagation even on timeout
-timeout --preserve-status .1 sleep 10
 # exit status should be 128+TERM
-test $? = 124 && fail=1
+returns_ 124 timeout --preserve-status .1 sleep 10 && fail=1
 
 # kill delay. Note once the initial timeout triggers,
 # the exit status will be 124 even if the command
 # exits on its own accord.
-timeout -s0 -k1 .1 sleep 10
-test $? = 124 && fail=1
+# exit status should be 128+KILL
+returns_ 124 timeout -s0 -k1 .1 sleep 10 && fail=1
+# Ensure a consistent exit status with --foreground
+returns_ 124 timeout --foreground -s0 -k1 .1 sleep 10 && fail=1
 
 # Ensure 'timeout' is immune to parent's SIGCHLD handler
 # Use a subshell and an exec to work around a bug in FreeBSD 5.0 /bin/sh.
 (
-  # ash doesn't support "trap '' CHLD"; it knows only signal numbers.
-  sig=$("$abs_top_builddir/src/kill" -l CHLD 2>/dev/null) && trap '' $sig
+  trap '' CHLD
 
   exec timeout 10 true
 ) || fail=1
@@ -61,4 +60,15 @@ out=$(sleep .1 & exec timeout .5 sh -c 'sleep 2; echo foo')
 status=$?
 test "$out" = "" && test $status = 124 || fail=1
 
+# Verify --verbose output
+cat > exp <<\EOF
+timeout: sending signal EXIT to command 'sleep'
+timeout: sending signal KILL to command 'sleep'
+EOF
+for opt in -v --verbose; do
+  timeout $opt -s0 -k .1 .1 sleep 10 2> errt
+  sed '/^Killed/d' < errt > err || framework_failure_
+  compare exp err || fail=1
+done
+
 Exit $fail
