@@ -2,7 +2,7 @@
 # Ensure that 'rm -d -i dir' (i.e., without --recursive) gives a prompt and
 # then deletes the directory if it is empty
 
-# Copyright (C) 2012-2013 Free Software Foundation, Inc.
+# Copyright (C) 2012-2023 Free Software Foundation, Inc.
 
 # This program is free software: you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -15,7 +15,7 @@
 # GNU General Public License for more details.
 
 # You should have received a copy of the GNU General Public License
-# along with this program.  If not, see <http://www.gnu.org/licenses/>.
+# along with this program.  If not, see <https://www.gnu.org/licenses/>.
 
 . "${srcdir=.}/tests/init.sh"; path_prepend_ ./src
 print_ver_ rm
@@ -28,7 +28,7 @@ printf "%s" \
     > exp.err || framework_failure_
 
 printf "%s\n" \
-    "removed directory: 'd'" \
+    "removed directory 'd'" \
     > exp || framework_failure_
 
 compare exp out || fail=1
